layouts = { };
elements = { };

baseDurationImageID = 0x41007E35;

categoryColor = { };
categoryColor[Turbine.Gameplay.EffectCategory.Fear] = Turbine.UI.Color(1.0, 1.0, 0.0, 1.0);
categoryColor[Turbine.Gameplay.EffectCategory.Wound] = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Disease] = Turbine.UI.Color(1.0, 1.0, 1.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Poison] = Turbine.UI.Color(1.0, 0.0, 1.0, 0.0);

PMCBackground = class();

function PMCBackground:Constructor(window, settings)
	self.window = window;
	self.opacity = settings.opacity;
	self.controls = { };
end

function PMCBackground:SetColor(r, g, b)
	self.window:SetBackColor(Turbine.UI.Color(self.opacity, r, g, b));

	for _, v in ipairs(self.controls) do
		v:SetBackColor(Turbine.UI.Color(v.opacity, r, g, b));
	end
end

PMCName = class(Turbine.UI.Label);

function PMCName:Constructor(window, settings)
	Turbine.UI.Label.Constructor(self);

	self:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self:SetFontStyle(Turbine.UI.FontStyle.Outline);
	self:SetFont(fontsize("Verdana", settings.h * 0.75));
	self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self:SetMultiline(false);
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetSize(settings.w, settings.h);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCName:SetColor(r, g, b)
	if r == nil then
		self:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	else
		self:SetForeColor(Turbine.UI.Color(1.0, r, g, b));
	end
end

function PMCName:SetName(name)
	self:SetText(name);
end

PMCMorale = class(moebius92.ColorBar.ColorBar);

function PMCMorale:Constructor(window, settings)
	moebius92.ColorBar.ColorBar.Constructor(self, 5, settings.w, settings.h);

	self.window = window;
	self.threat = 0;
	self.damage = 0;
	self.damageLog = { };
	self:SetBarColor(1, Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, 1.0));
	self:SetBarColor(3, Turbine.UI.Color(1.0, 1.0, 0.5, 0.0));
	self:SetBarColor(4, Turbine.UI.Color(1.0, 0.5, 1.0, 0.5));
	self:SetBarColor(5, Turbine.UI.Color(1, 1.0, 0.0, 0.0));
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCMorale:SetMorale(cur, base, max, temp, tmax, gmax, old, combat, tend)
	local m = math.max(max, base) + temp;
	local min = math.min(max, base);
	local damage = old - cur;

	if combat then
		local t = self.window.parent.options.controls[self.index].threat;
		if t > 0 and damage > 0 then
			local time = math.floor(Turbine.Engine.GetGameTime());

			self.damage = self.damage + damage;
			if self.damageLog[time] == nil then
				self.damageLog[time] = damage;
				for k, v in pairs(self.damageLog) do
					if (time - k) >= t then
						self.damageLog[k] = nil;
						self.damage = self.damage - v;
					end
				end
			else
				self.damageLog[time] = self.damageLog[time] + damage;
			end

			if self.threat < self.damage then
				self.threat = self.damage;
			end
		end
	else
		self.threat = 0;
		self.damage = 0;
	end

	self.tempEnd = tend;
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, 1.0));
	if self.tempEnd ~= 0 then
		self:SetWantsUpdates(true);
		self:Update();
	else
		self:SetWantsUpdates(false);
	end

	self:SetStats(1, 0, cur, min, gmax);
	self:SetStats(2, max, max + temp, max + temp, gmax);
	self:SetStats(3, max + temp, m, m, gmax);
	self:SetStats(4, base, cur, max, gmax);
	self:SetStats(5, 0, cur, math.min(self.threat, max), gmax);
end

function PMCMorale:Update()
	if self.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, t));
end

PMCMoraleNG = class(moebius92.ColorBar.ColorBar);

function PMCMoraleNG:Constructor(window, settings)
	moebius92.ColorBar.ColorBar.Constructor(self, 4, settings.w, settings.h);

	self.window = window;
	self:SetBarColor(1, Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, 1.0));
	self:SetBarColor(3, Turbine.UI.Color(1.0, 1.0, 0.5, 0.0));
	self:SetBarColor(4, Turbine.UI.Color(1.0, 0.5, 1.0, 0.5));
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCMoraleNG:SetMorale(cur, base, max, temp, tmax, gmax, old, combat, tend)
	local m = math.max(max, base) + temp;
	local min = math.min(max, base);
	local damage = old - cur;

	self.tempEnd = tend;
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, 1.0));
	if self.tempEnd ~= 0 then
		self:SetWantsUpdates(true);
		self:Update();
	else
		self:SetWantsUpdates(false);
	end

	self:SetStats(1, 0, cur, min, m);
	self:SetStats(2, max, max + temp, max + temp, m);
	self:SetStats(3, max + temp, m, m, m);
	self:SetStats(4, base, cur, max, m);
end

function PMCMoraleNG:SetColor(r, g, b)
	if r == nil then
		self:SetBarColor(1, Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
	else
		self:SetBarColor(1, Turbine.UI.Color(1.0, r, g, b));
	end
end

function PMCMoraleNG:Update()
	if self.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 0.0, t));
end

PMCPower = class(moebius92.ColorBar.ColorBar);

function PMCPower:Constructor(window, settings)
	moebius92.ColorBar.ColorBar.Constructor(self, 4, settings.w, settings.h);

	self.window = window;

	self:SetBarColor(1, Turbine.UI.Color(1.0, 0.0, 0.0, 1.0));
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 1.0, 0.0));
	self:SetBarColor(3, Turbine.UI.Color(1.0, 1.0, 0.0, 0.5));
	self:SetBarColor(4, Turbine.UI.Color(1.0, 0.5, 0.5, 1.0));
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCPower:SetPower(cur, base, max, temp, tmax, gmax, old, combat, tend)
	local m = math.max(max, base) + temp;

	self.tempEnd = tend;
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 1.0, 0.0));
	if self.tempEnd ~= 0 then
		self:SetWantsUpdates(true);
		self:Update();
	else
		self:SetWantsUpdates(false);
	end

	self:SetStats(1, 0, cur, max, gmax);
	self:SetStats(2, max, max + temp, max + temp, gmax);
	self:SetStats(3, max + temp, m, m, gmax);
	self:SetStats(4, math.max(0, max - self.window.parent.options.controls[self.index].fill), cur, max, gmax);
end

function PMCPower:Update()
	if self.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, t, 0.0));
end

PMCPowerNG = class(moebius92.ColorBar.ColorBar);

function PMCPowerNG:Constructor(window, settings)
	moebius92.ColorBar.ColorBar.Constructor(self, 4, settings.w, settings.h);

	self.window = window;

	self:SetBarColor(1, Turbine.UI.Color(1.0, 0.0, 0.0, 1.0));
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 1.0, 0.0));
	self:SetBarColor(3, Turbine.UI.Color(1.0, 1.0, 0.0, 0.5));
	self:SetBarColor(4, Turbine.UI.Color(1.0, 0.5, 0.5, 1.0));
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCPowerNG:SetPower(cur, base, max, temp, tmax, gmax, old, combat, tend)
	local m = math.max(max, base) + temp;

	self.tempEnd = tend;
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, 1.0, 0.0));
	if self.tempEnd ~= 0 then
		self:SetWantsUpdates(true);
		self:Update();
	else
		self:SetWantsUpdates(false);
	end

	self:SetStats(1, 0, cur, max, m);
	self:SetStats(2, max, max + temp, max + temp, m);
	self:SetStats(3, max + temp, m, m, m);
	self:SetStats(4, math.max(0, max - self.window.parent.options.controls[self.index].fill), cur, max, m);
end

function PMCPowerNG:Update()
	if self.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self:SetBarColor(2, Turbine.UI.Color(1.0, 0.5, t, 0.0));
end

PMCSquare = class(Turbine.UI.Control);

function PMCSquare:Constructor(window, settings)
	Turbine.UI.Control.Constructor(self);

	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetSize(settings.w, settings.h);
	self:SetParent(window);
	self:SetMouseVisible(false);
end

function PMCSquare:SetColor(r, g, b)
	if r == nil then
		self:SetVisible(false);
	else
		self:SetVisible(true);
		self:SetBackColor(Turbine.UI.Color(1.0, r, g, b));
	end
end

PMCSquareOutline = class(Turbine.UI.Control);

function PMCSquareOutline:Constructor(window, settings)
	Turbine.UI.Control.Constructor(self);

	self.window = window;
	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetSize(settings.w, settings.h);
	self:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self:SetParent(window);
	self:SetMouseVisible(false);
	self.control = Turbine.UI.Control();
	self.control:SetPosition(1, 1);
	self.control:SetSize(settings.w - 2, settings.h - 2);
	self.control:SetParent(self);
	self.control:SetMouseVisible(false);
end

function PMCSquareOutline:SetColor(r, g, b)
	self.r = r;
	self.g = g;
	self.b = b;

	if r == nil then
		self:SetVisible(false);
	else
		self:SetVisible(true);
		self.control:SetBackColor(Turbine.UI.Color(1.0, r, g, b));
	end
end

function PMCSquareOutline:SetEffect(e)
	if e ~= nil then
		self.durationTime = e:GetDuration();
		self.startTime = e:GetStartTime();
		if self.durationTime < 1000000 then
			self:SetWantsUpdates(true);
			self:Update();
			return;
		end
	end

	self.durationTime = nil;
	self.startTime = nil;
	self:SetWantsUpdates(false);
end

function PMCSquareOutline:Update()
	local time = Turbine.Engine.GetGameTime();
	local x = self.window.parent.options.controls[self.index].effectDuration;

	if x > 0 then
		x = 1 - math.min(0.5, math.max(0, 0.5 * (time - (self.startTime + self.durationTime - x)) / x));
	else
		x = 1 - math.min(0.5, math.max(0, 0.5 * (time - self.startTime) / self.durationTime));
	end

	if self.r == nil then
		self:SetWantsUpdates(false);
	else
		self.control:SetBackColor(Turbine.UI.Color(1.0, self.r * x, self.g * x, self.b * x));
	end
end

PMCEffect = class(Turbine.UI.Control);

function PMCEffect:Constructor(window, settings)
	Turbine.UI.Control.Constructor(self);

	self.window = window;
	self.w = settings.w;
	self.h = settings.h;

	self:SetPosition(settings.x, settings.y);
	self:SetZOrder(settings.z);
	self:SetSize(settings.w, settings.h);
	self:SetVisible(true);
	self:SetMouseVisible(false);
	self:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self:SetParent(window);
	self.display = Turbine.UI.Lotro.EffectDisplay();
	self.display:SetPosition(0, 0);
	self.display:SetZOrder(settings.z + 1);
	self.display:SetSize(settings.w, settings.h);
	self.display:SetVisible(false);
	self.display:SetParent(self);
	self.duration = Turbine.UI.Control();
	self.duration:SetPosition(0, 0);
	self.duration:SetSize(settings.w, settings.h);
	self.duration:SetZOrder(settings.z + 2);
	self.duration:SetOpacity(0.7);
	self.duration:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
	self.duration:SetMouseVisible(false);
	self.duration:SetParent(self.display);
	self.text = Turbine.UI.Label();
	self.text:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.text:SetFontStyle(Turbine.UI.FontStyle.Outline);
	self.text:SetFont(fontsize("Verdana", settings.h * 0.6));
	self.text:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
	self.text:SetPosition(0, 0);
	self.text:SetSize(settings.w, settings.h);
	self.text:SetZOrder(settings.z + 3);
	self.text:SetStretchMode(1);
	self.text:SetMouseVisible(false);
	self.text:SetParent(self.display);
end

function PMCEffect:SetColor(r, g, b)
	if r == nil then
		self:SetBackColor(Turbine.UI.Color(0.0, 0.0, 0.0, 0.0));
		self.duration:SetBackColor(Turbine.UI.Color(0.0, 0.0, 0.0, 0.0));
	else
		self:SetBackColor(Turbine.UI.Color(0.5, r, g, b));
		self.duration:SetBackColor(Turbine.UI.Color(1.0, r, g, b));
	end
end

function PMCEffect:SetIcon(i)
	self.display:SetVisible(false);
	self:SetWantsUpdates(false);

	if i == nil then
		self:SetBackground(nil);
		self:SetVisible(false);
		self:SetVisible(true);
	else
		self:SetBackground(i);
		self:SetStretchMode(2);
		self:SetStretchMode(1);
		self:SetSize(self.w, self.h);
	end
end

function PMCEffect:SetEffect(e)
	self:SetBackground(nil);
	self:SetVisible(false);
	self:SetVisible(true);

	if e == nil then
		self.display:SetVisible(false);
	else
		self.display:SetVisible(true);
		self.display:SetEffect(e);

		self.durationTime = e:GetDuration();
		self.startTime = e:GetStartTime();
		if self.durationTime < 1000000 then
			self.duration:SetVisible(true);
			self:SetWantsUpdates(true);
			self:Update();
		else
			self.duration:SetVisible(false);
			self:SetWantsUpdates(false);
		end

		local cat = e:GetCategory();
		local c = categoryColor[cat];
		local n = 0;
		if c ~= nil and e:IsDebuff() and e:IsCurable() then
			for _, v in ipairs(self.effects) do
				if v.item.effect:IsDebuff() and v.item.effect:IsCurable() and v.item.effect:GetCategory() == cat then
					n = n + 1;
				end
			end

			if n > 1 then
				self.text:SetText(tostring(n));
				self.text:SetForeColor(c);
				self.text:SetVisible(true);
			else
				self.text:SetVisible(false);
			end
		else
			local name = e:GetName();
			local t = string.match(name, " ([0-9]+)$") or string.match(name, "^Tier ([0-9]+)");
			n = #self.effects;

			if t ~= nil then
				self.text:SetText(t);
				self.text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
				self.text:SetVisible(true);
			elseif n > 1 then
				self.text:SetText(tostring(n));
				self.text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
				self.text:SetVisible(true);
			else
				self.text:SetVisible(false);
			end
		end
	end
end

function PMCEffect:Update()
	local time = Turbine.Engine.GetGameTime();
	local x = self.window.parent.options.controls[self.index].effectDuration;

	if x > 0 then
		x = math.floor(math.min(59, math.max(0, 60.0 * (time - (self.startTime + self.durationTime - x)) / x)));
	else
		x = math.floor(math.min(59, math.max(0, 60.0 * (time - self.startTime) / self.durationTime)));
	end

	if self.durationIcon ~= x then
		self.duration:SetBackground(baseDurationImageID + x);
		self.duration:SetStretchMode(2);
		self.duration:SetStretchMode(1);
		self.duration:SetSize(self.w, self.h);
		self.durationIcon = x;
	end
end