PartyMemberWindow = class(Turbine.UI.Window);

function PartyMemberWindow:Constructor(parent, actor, layout, petclass)
	Turbine.UI.Window.Constructor(self);

	self.parent = parent;
	self.actor = actor;
	self.petclass = petclass;
	if self.petclass == nil then
		self.class = actor:GetClass();
	else
		self.class = petclass;
	end

	self.rolls = { };
	self.rollOrder = nil;
	self.actorHandle = { };

	if layout.pre ~= nil then
		layout.pre(self);
	end

	self:SetSize(layout.width, layout.height);
	self.controls = { };
	for k, v in ipairs(layout.layout) do
		if v.element ~= nil then
			self.controls[k] = v.element(self, v.settings);
		else
			self.controls[k] = { };
		end
		self.controls[k].index = k;
	end

	if layout.post ~= nil then
		layout.post(self);
	end

	self.select = Turbine.UI.Lotro.EntityControl();
	self.select:SetParent(self);
	self.select:SetPosition(0, 0);
	self.select:SetMouseVisible(true);
	self.select:SetEntity(actor);
	self.select:SetSelectionEnabled(true);
	self.select:SetContextMenuEnabled(true);
	self.select:SetSize(self:GetSize());
	self:SetMouseVisible(true);

	self.actorHandle["NameChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "NameChanged", function(sender, args) self:UpdateName(); end);

	self.oldMorale = actor:GetMorale();
	self.tempMoraleEffect = { };
	self.tempMoraleEnd = 0;
	self.actorHandle["MoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["BaseMaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["TemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxTemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);

	self.oldPower = actor:GetPower();
	self.tempPowerEffect = { };
	self.tempPowerEnd = 0;
	self.actorHandle["PowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["BaseMaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["TemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxTemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);

	self.effects = self.actor:GetEffects();
	self.effectsHandle = { };
	self.effectsHandle["EffectAdded"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:AddEffect(self.effects:Get(args.Index)); end);
	self.effectsHandle["EffectRemoved"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectRemoved", function(sender, args) self:RemoveEffect(args.Effect, true); end);
	self.effectsHandle["EffectsCleared"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectsCleared", function(sender, args) self:ClearEffects(); end);

	if self.petclass == nil then
		self.actorHandle["IsLinkDeadChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsLinkDeadChanged", function(sender, args) self:SetStatus("LinkDead", self.actor:IsLinkDead()); end);
		self.actorHandle["IsVoiceEnabledChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceEnabledChanged", function(sender, args) self:SetStatus("Voice", self.actor:IsVoiceEnabled()); end);
		self.actorHandle["IsVoiceActiveChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceActiveChanged", function(sender, args) self:SetStatus("Talking", self.actor:IsVoiceActive()); end);
		self.actorHandle["ReadyStateChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "ReadyStateChanged", function(sender, args) self:SetStatus("Ready", self.actor:GetReadyState()); end);
		self.actorHandle["PetChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PetChanged", function(sender, args) self.parent:PartyMemberChanged(); end);
	end

	self:InitializeControls();

	self:SetVisible(self.parent.hide);
end

function PartyMemberWindow:RemoveHandlers()
	for k, v in pairs(self.actorHandle) do
		moebius92.Handlers.RemoveEventHandler(self.actor, k, v);
	end
	for k, v in pairs(self.effectsHandle) do
		moebius92.Handlers.RemoveEventHandler(self.effects, k, v);
	end
end

function PartyMemberWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = true;
		self.safeY = nil;
	end
end

function PartyMemberWindow:MouseMove(args)
	local w, h = self:GetSize();

	if self.dragging then
		local x, y;

		if not IsSafeY(self, args.Y) then
			return;
		end

		x = math.floor(args.X / w);
		y = math.floor(args.Y / h);
		if x == 0 and y == 1 and self.parent.arrange == 1 and self.pet ~= nil then
			y = 0;
		end

		if x ~= 0 or y ~= 0 then
			if self.parent:MoveMember(self.id, x, y) and self.parent.arrange == 1 then
				if y < 0 then
					self.safeY = math.huge;
				else
					self.safeY = -math.huge;
				end
			end
		end
	end
end

function PartyMemberWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = false;
	end
end

function PartyMemberWindow:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		self.parent.memberSelect = self;
		for i = 1, #self.parent.memberOptions, 1 do
			self.parent.memberOptions[i]:SetChecked(i == self.showPet);
			if self.parent.arrange == 1 then
				self.parent.memberOptions[i]:SetEnabled(i <= 2);
			else
				self.parent.memberOptions[i]:SetEnabled(true);
			end
		end
		self.parent.memberMenu:ShowMenu();
	end
end

function PartyMemberWindow:UpdateName()
	local name = self.actor:GetName();

	if name == nil or name == "" then
		self.pollName = true;
		self:SetWantsUpdates(true);
	else
		self.pollName = false;
	end

	if self.rollOrder ~= nil then
		name = "(" .. self.rollOrder .. ")" .. name;
	end

	for _, v in pairs(self.controls) do
		if v.SetName ~= nil then
			v:SetName(name);
		end
	end
end

function PartyMemberWindow:UpdateMorale()
	local cur = self.actor:GetMorale();
	local base = self.actor:GetBaseMaxMorale();
	local max = self.actor:GetMaxMorale();
	local temp = self.actor:GetTemporaryMorale();
	local tmax = self.actor:GetMaxTemporaryMorale();

	local s = self.parent:MoraleControl(cur, cur / max);
	local t = { };
	for k, v in pairs(self.moraleState) do
		if s[k] ~= v then
			t[k] = v;
		end
	end
	self:DecrementControlGroup(t);
	t = { };
	for k, v in pairs(s) do
		if self.moraleState[k] ~= v then
			t[k] = v;
		end
	end
	self:IncrementControlGroup(t);
	self.moraleState = s;

	for _, v in pairs(self.controls) do
		if v.SetMorale ~= nil then
			v:SetMorale(cur, base, max, temp, tmax, self.parent.maxMorale, self.oldMorale, self.parent.inCombat, self.tempMoraleEnd);
		end
	end

	self.maxMorale = math.max(max, base) + math.min(tmax, base);
	self.oldMorale = cur;
	self.parent:UpdateMorale();
end

function PartyMemberWindow:UpdatePower()
	local cur = self.actor:GetPower();
	local base = self.actor:GetBaseMaxPower();
	local max = self.actor:GetMaxPower();
	local temp = self.actor:GetTemporaryPower();
	local tmax = self.actor:GetMaxTemporaryPower();

	local s = self.parent:PowerControl(cur, cur / max);
	local t = { };
	for k, v in pairs(self.powerState) do
		if s[k] ~= v then
			t[k] = v;
		end
	end
	self:DecrementControlGroup(t);
	t = { };
	for k, v in pairs(s) do
		if self.powerState[k] ~= v then
			t[k] = v;
		end
	end
	self:IncrementControlGroup(t);
	self.powerState = s;

	for _, v in pairs(self.controls) do
		if v.SetPower ~= nil then
			v:SetPower(cur, base, max, temp, tmax, self.parent.maxPower, self.oldPower, self.parent.inCombat, self.tempPowerEnd);
		end
	end

	self.maxPower = math.max(max, base) + math.min(tmax, base);
	self.oldPower = cur;
	self.parent:UpdatePower();
end

function PartyMemberWindow:SetStatus(status, v)
	if v == nil or self.statusState[status] ~= v then
		local t;

		if self.statusState[status] ~= nil then
			t = self.parent:StatusControl(status, self.statusState[status]);
			self:DecrementControlGroup(t);
		end

		t = self.parent:StatusControl(status, v);
		self:IncrementControlGroup(t);

		self.statusState[status] = v;
	end
end

function PartyMemberWindow:AddEffect(effect)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	self.parent:LogEffect(effect);

	self.lastEffectUpdate = time;
	for i = 1, self.effects:GetCount(), 1 do
		id = self.effects:Get(i):GetID();

		if self.effectState[id] ~= nil then
			self.effectState[id].time = time;
		end
	end
	self:SetWantsUpdates(true);

	id = effect:GetID();

	if tempMoraleEffects[name] then
		self.tempMoraleEffect[id] = effect:GetStartTime() + effect:GetDuration();
		self.tempMoraleEnd = 0;
		for _, v in pairs(self.tempMoraleEffect) do
			self.tempMoraleEnd = math.max(self.tempMoraleEnd, v);
		end
		self:UpdateMorale();
	end

	if self.effectState[id] == nil then
		local t = self.parent:EffectControl(effect:IsDebuff(), effect:GetIcon(), name, effect:GetCategory(), effect:IsCurable());

		if next(t) ~= nil then
			local item = { };

			item.effect = effect;
			item.time = time;
			item.controls = t;
			item.startTime = effect:GetStartTime();
			item.durationTime = effect:GetDuration();
			item.endTime = item.startTime + item.durationTime;
			self.effectState[id] = item;

			self:IncrementControlGroup(t);
			for k, v in pairs(t) do
				table.insert(self.controls[k].effects, { item = item, group = v });
				table.sort(self.controls[k].effects, function(a, b) return a.group < b.group or (a.group == b.group and a.item.startTime > b.item.startTime); end);
				self:UpdateControlEffect(k);
			end
		end
	end
end

function PartyMemberWindow:RemoveEffect(effect, mark)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	if mark then
		self.lastEffectUpdate = time;
		for i = 1, self.effects:GetCount(), 1 do
			id = self.effects:Get(i):GetID();

			if self.effectState[id] ~= nil then
				self.effectState[id].time = time;
			end
		end
	end

	id = effect:GetID();

	if self.tempMoraleEffect[id] ~= nil then
		self.tempMoraleEffect[id] = nil;
		self.tempMoraleEnd = 0;
		for _, v in pairs(self.tempMoraleEffect) do
			self.tempMoraleEnd = math.max(self.tempMoraleEnd, v);
		end
	end

	if self.effectState[id] ~= nil then
		self:DecrementControlGroup(self.effectState[id].controls);
		for k in pairs(self.effectState[id].controls) do
			for k1, v1 in ipairs(self.controls[k].effects) do
				if v1.item == self.effectState[id] then
					table.remove(self.controls[k].effects, k1);
					self:UpdateControlEffect(k);
					break;
				end
			end
		end
		self.effectState[id] = nil;
	end
end

function PartyMemberWindow:ClearEffects()
	for _, v in pairs(self.effectState[id]) do
		self:DecrementControlGroup(self.effectState[id].controls);
	end
	self.effectState = { };

	for k, v in ipairs(self.controls) do
		self.effects = { };
		self:UpdateControlEffect(k);
	end

	self.tempMoraleEffect = { };
	self.tempMoraleEnd = 0;
	self:UpdateMorale();

	self.tempPowerEffect = { };
	self.tempPowerEnd = 0;
	self:UpdatePower();
end

function PartyMemberWindow:CheckEffects(t)
	for k, v in pairs(self.effectState) do
		if not t[k] then
			self:RemoveEffect(v.effect, false);
		end
	end
end

function PartyMemberWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local update = false;

	if self.lastEffectUpdate == time then
		for k, v in pairs(self.effectState) do
			if v.time ~= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	else
		for k, v in pairs(self.effectState) do
			if v.endTime <= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	end

	for k, v in pairs(self.effectState) do
		if v.durationTime < 1000000 then
			update = true;
			break;
		end
	end

	if self.pollName then
		self:UpdateName();
		update = true;
	end

	self:SetWantsUpdates(update);
end

function PartyMemberWindow:InitializeControls()
	self.effectState = { };
	self.statusState = { };
	self.moraleState = { };
	self.powerState = { };
	for _, v in ipairs(self.controls) do
		v.effects = { };
		v.group = { };
	end

	self:UpdateName();
	self:UpdateMorale();
	self:UpdatePower();

	self.lastEffectUpdate = 0;
	for i = 1, self.effects:GetCount(), 1 do
		self:AddEffect(self.effects:Get(i));
	end

	if self.petclass == nil then
		self:SetStatus("LinkDead", self.actor:IsLinkDead());
		self:SetStatus("Voice", self.actor:IsVoiceEnabled());
		self:SetStatus("Talking", self.actor:IsVoiceActive());
		self:SetStatus("Ready", self.actor:GetReadyState());
	end
	self:SetStatus("Default", nil);
	if self.petclass == nil then
		self:SetStatus("Player", nil);
	else
		self:SetStatus("Pet", nil);
	end
	for k, v in pairs(Turbine.Gameplay.Class) do
		if v == self.class then
			self:SetStatus(k, nil);
			break;
		end
	end

	for k in ipairs(self.controls) do
		self:UpdateControlColor(k);
		self:UpdateControlEffect(k);
	end
end

function PartyMemberWindow:IncrementControlGroup(t)
	for k, v in pairs(t) do
		if self.controls[k].group[v] == nil then
			self.controls[k].group[v] = 0;
			self:UpdateControlColor(k);
		end
		self.controls[k].group[v] = self.controls[k].group[v] + 1;
	end
end

function PartyMemberWindow:DecrementControlGroup(t)
	for k, v in pairs(t) do
		self.controls[k].group[v] = self.controls[k].group[v] - 1;
		if self.controls[k].group[v] <= 0 then
			self.controls[k].group[v] = nil;
			self:UpdateControlColor(k);
		end
	end
end

function PartyMemberWindow:UpdateControlColor(i)
	local controlOption = self.parent.options.controls[i];

	if self.controls[i].SetColor ~= nil and controlOption.color ~= nil then
		local H, S, V;

		for k, v in ipairs(controlOption.color) do
			if controlOption.missing == (self.controls[i].group[k] == nil) then
				if v.color.V ~= nil then
					if V == nil then
						V = v.color.V;
					else
						V = (V + v.color.V) / 2;
					end
				end

				if v.color.H ~= nil then
					if H == nil then
						H = v.color.H;
						S = v.color.S;
					else
						local r = v.color.S / 2;
						local theta = (math.acos(-r / 2) - 2 * math.pi / 3) * S + (2 * math.pi / 3);
						local H1 = (v.color.H - H) % 6;

						if H1 <= 2 then
							theta = theta * H1 / 2;
						elseif H1 <= 4 then
							theta = theta + theta * (H1 - 2);
						else
							theta = -theta * (6 - H1) / 2;
						end

						local x = math.cos(theta) * r + S;
						local y = math.sin(theta) * r;
						r = math.sqrt(x * x + y * y);
						if r == 0 then
							H = 0.0;
							S = 0.0;
						else
							H = ((math.atan(y / r, x / r) * 3 / math.pi) + H) % 6;
							S = math.min(r, 1.0);
						end
					end
				end

				if v.color.H ~= nil and v.color.V ~= nil then
					break;
				end
			end
		end

		if H == nil then
			self.controls[i]:SetColor(nil, nil, nil);
		else
			local c = { };

			c[0] = 1.0 - S;
			c[1] = 1.0 - S
			c[2] = 1.0 - S
			H = (H + 1) % 6;
			x = math.floor(H / 2);
			c[x] = c[x] + S;
			H = H % 2;
			if H < 1 then
				x = (x - 1) % 3;
				c[x] = c[x] + S * (1.0 - H);
			else
				x = (x + 1) % 3;
				c[x] = c[x] + S * (H - 1.0);
			end
			V = V and V or 1.0;
			self.controls[i]:SetColor(c[0] * V, c[1] * V, c[2] * V);
		end
	end
end

function PartyMemberWindow:UpdateControlEffect(i)
	if self.parent.options.controls[i].missing then
		if self.controls[i].SetIcon ~= nil then
			if self.parent.options.controls[i].color ~= nil then
				local icon = nil;

				for k, v in ipairs(self.parent.options.controls[i].color) do
					if self.controls[i].group[k] == nil then
						icon = self.parent.options.controls[i].color[k][1].icon;
						break;
					end
				end

				if icon == 0 then
					icon = nil;
				end

				self.controls[i]:SetIcon(icon);
			else
				self.controls[i]:SetIcon(nil);
			end
		end
	else
		if self.controls[i].SetEffect ~= nil then
			if #self.controls[i].effects > 0 then
				self.controls[i]:SetEffect(self.controls[i].effects[1].item.effect);
			else
				self.controls[i]:SetEffect(nil);
			end
		elseif self.controls[i].SetIcon ~= nil then
			if #self.controls[i].effects > 0 then
				self.controls[i]:SetIcon(self.controls[i].effects[1].item.effect:GetIcon());
			else
				self.controls[i]:SetIcon(nil);
			end
		end
	end
end